@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -41,6 +41,7 @@
 #include <connman/option.h>
 #include <connman/inet.h>
 #include <connman/dbus.h>
+#include <connman/wifi.h>
 #include <connman/log.h>
 
 #include "supplicant.h"
@@ -174,15 +175,24 @@ struct supplicant_result {
 	dbus_int32_t maxrate;
 };
 
+struct supplicant_block {
+	unsigned char *ssid;
+	char *netpath;
+	gboolean enabled;
+	int num_scans;
+};
+
 struct supplicant_task {
 	int ifindex;
 	char *ifname;
-	gboolean mac80211;
+	gboolean cfg80211;
 	struct connman_device *device;
 	struct connman_network *network;
 	struct connman_network *pending_network;
 	char *path;
 	char *netpath;
+	gboolean hidden_found;
+	GHashTable *hidden_blocks;
 	gboolean created;
 	enum supplicant_state state;
 	gboolean scanning;
@@ -206,6 +216,16 @@ static void free_task(struct supplicant_task *task)
 	g_free(task);
 }
 
+static void remove_block(gpointer user_data)
+{
+	struct supplicant_block *block = user_data;
+
+	DBG("");
+
+	g_free(block->ssid);
+	g_free(block->netpath);
+}
+
 static struct supplicant_task *find_task_by_index(int index)
 {
 	GSList *list;
@@ -303,18 +323,270 @@ static int get_bssid(struct connman_device *device,
 	return 0;
 }
 
+static int enable_network(struct supplicant_task *task, const char *netpath,
+			  connman_bool_t enable)
+{
+	DBusMessage *message, *reply;
+	DBusError error;
+	char *enable_string;
+
+	DBG("enable %d", enable);
+
+	enable_string = enable ? "enable" : "disable";
+
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, netpath,
+				SUPPLICANT_INTF ".Network", enable_string);
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_set_auto_start(message, FALSE);
+
+	dbus_error_init(&error);
+
+	reply = dbus_connection_send_with_reply_and_block(connection,
+							message, -1, &error);
+	if (reply == NULL) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Failed to select network");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
+	dbus_message_unref(reply);
+
+	dbus_message_unref(message);
+
+	return 0;
+}
+
+static int set_hidden_network(struct supplicant_task *task, const char *netpath,
+				const unsigned char *ssid, int ssid_len)
+{
+	DBusMessage *message, *reply;
+	DBusMessageIter array, dict;
+	DBusError error;
+	dbus_uint32_t scan_ssid = 1;
+	const char *invalid_address = "ff:ff:ff:ff:ff:ff";
+
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, netpath,
+					SUPPLICANT_INTF ".Network", "set");
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_set_auto_start(message, FALSE);
+
+	dbus_message_iter_init_append(message, &array);
+
+	connman_dbus_dict_open(&array, &dict);
+
+	connman_dbus_dict_append_basic(&dict, "scan_ssid",
+					 DBUS_TYPE_UINT32, &scan_ssid);
+
+	connman_dbus_dict_append_fixed_array(&dict, "ssid",
+					DBUS_TYPE_BYTE, &ssid, ssid_len);
+
+	/*
+	 * We're setting an invalid BSSID to prevent wpa_s from associating
+	 * automatically to this block once it's found.
+	 */
+	connman_dbus_dict_append_basic(&dict, "bssid",
+					DBUS_TYPE_STRING, &invalid_address);
+
+	connman_dbus_dict_close(&array, &dict);
+
+	dbus_error_init(&error);
+
+	reply = dbus_connection_send_with_reply_and_block(connection,
+							message, -1, &error);
+	if (reply == NULL) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Failed to set network options");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
+	dbus_message_unref(reply);
+
+	dbus_message_unref(message);
+
+	return 0;
+}
+
+static void block_reset(gpointer key, gpointer value, gpointer user_data)
+{
+	struct supplicant_block *block = value;
+	struct supplicant_task *task = user_data;
+
+	block->num_scans = 0;
+	if (block->enabled)
+		enable_network(task, block->netpath, FALSE);
+
+	block->enabled = FALSE;
+}
+
+#define MAX_BLOCK_SCANS 2
+static void hidden_block_enable(struct supplicant_task *task)
+{
+	GHashTableIter iter;
+	gpointer key, value;
+	struct supplicant_block *block;
+
+	DBG("network %p", task->network);
+
+	if (g_hash_table_size(task->hidden_blocks) == 0)
+		return;
+
+	/*
+	 * If we're associated or associating, we no longer need to
+	 * look for hidden networks.
+	 */
+	if (task->network)
+		return;
+
+	/*
+	 * We go through the block list and:
+	 * - If we scanned it more than twice, we disable it and move
+	 *   on to the next block.
+	 * - If the next block is not enabled, we enable it, start
+	 *   the scan counter, and return. This routine will be called
+	 *   again when the next scan results are available.
+	 * - If we're done with all the blocks there, we just reset them.
+	 */
+	g_hash_table_iter_init(&iter, task->hidden_blocks);
+	while (g_hash_table_iter_next(&iter, &key, &value)) {
+		block = value;
+
+		DBG("%s num of scans %d enabled %d",
+			block->ssid, block->num_scans, block->enabled);
+
+		if (block->num_scans > MAX_BLOCK_SCANS) {
+			if (block->enabled == FALSE)
+				continue;
+
+			enable_network(task, block->netpath, FALSE);
+			block->enabled = FALSE;
+			continue;
+		}
+
+		if (block->enabled == FALSE) {
+			enable_network(task, block->netpath, TRUE);
+			block->enabled = TRUE;
+		}
+
+		block->num_scans++;
+
+		return;
+	}
+
+	g_hash_table_foreach(task->hidden_blocks, block_reset, task);
+}
+
+static int add_hidden_network(struct supplicant_task *task,
+				const unsigned char *ssid, int ssid_len)
+{
+	DBusMessage *message, *reply;
+	DBusError error;
+	const char *path;
+	struct supplicant_block *block;
+	char *netpath = NULL;
+	int ret, i;
+
+	DBG("task %p", task);
+
+	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
+				SUPPLICANT_INTF ".Interface", "addNetwork");
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_set_auto_start(message, FALSE);
+
+	dbus_error_init(&error);
+
+	reply = dbus_connection_send_with_reply_and_block(connection,
+							message, -1, &error);
+	if (reply == NULL) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Failed to add network");
+		dbus_message_unref(message);
+		return -EIO;
+	}
+
+	dbus_error_init(&error);
+
+	if (dbus_message_get_args(reply, &error, DBUS_TYPE_OBJECT_PATH, &path,
+						DBUS_TYPE_INVALID) == FALSE) {
+		if (dbus_error_is_set(&error) == TRUE) {
+			connman_error("%s", error.message);
+			dbus_error_free(&error);
+		} else
+			connman_error("Wrong arguments for network");
+		dbus_message_unref(reply);
+		return -EIO;
+	}
+
+	netpath = g_strdup(path);
+
+	ret = set_hidden_network(task, netpath, ssid, ssid_len);
+	if (ret < 0)
+		goto done;
+
+	block = g_try_new0(struct supplicant_block, 1);
+	if (block == NULL)
+		goto done;
+
+	block->ssid = g_try_malloc0(ssid_len + 1);
+	if (block->ssid == NULL) {
+		g_free(block);
+		goto done;
+	}
+
+	for (i = 0; i < ssid_len; i++) {
+		if (g_ascii_isprint(ssid[i]))
+			block->ssid[i] = ssid[i];
+		else
+			block->ssid[i] = ' ';
+	}
+
+	block->netpath = netpath;
+	block->enabled = FALSE;
+	block->num_scans = 0;
+
+	DBG("path %s ssid %s", block->netpath, block->ssid);
+
+	g_hash_table_replace(task->hidden_blocks, block->ssid, block);
+
+	return 0;
+done:
+	g_free(netpath);
+
+	dbus_message_unref(reply);
+
+	dbus_message_unref(message);
+
+	return ret;
+}
+
 static void add_interface_reply(DBusPendingCall *call, void *user_data)
 {
 	struct supplicant_task *task = user_data;
 	DBusMessage *reply;
 	DBusError error;
 	const char *path;
+	char **hex_ssids, *hex_ssid;
+	int i;
 
 	DBG("task %p", task);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
 		goto failed;
@@ -340,11 +612,42 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 
 	dbus_message_unref(reply);
 
+	dbus_pending_call_unref(call);
+
+	hex_ssids = connman_wifi_load_ssid();
+	if (hex_ssids == NULL)
+		return;
+
+	for (i = 0; hex_ssids[i]; i++) {
+		unsigned char *ssid;
+		unsigned int j, k = 0, hex;
+		size_t hex_ssid_len;
+
+		hex_ssid = hex_ssids[i];
+		hex_ssid_len = strlen(hex_ssid);
+
+		ssid = g_try_malloc0(hex_ssid_len / 2 + 1);
+		if (ssid == NULL)
+			break;
+
+		for (j = 0, k = 0; j < hex_ssid_len; j += 2) {
+			sscanf(hex_ssid + j, "%02x", &hex);
+			ssid[k++] = hex;
+		}
+
+		if (add_hidden_network(task, ssid, hex_ssid_len / 2) < 0)
+			break;
+	}
+
+	g_strfreev(hex_ssids);
+
 	return;
 
 failed:
 	dbus_message_unref(reply);
 
+	dbus_pending_call_unref(call);
+
 	task_list = g_slist_remove(task_list, task);
 
 	connman_device_unref(task->device);
@@ -410,8 +713,6 @@ static void get_interface_reply(DBusPendingCall *call, void *user_data)
 	DBG("task %p", task);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
 		add_interface(task);
@@ -439,6 +740,8 @@ static void get_interface_reply(DBusPendingCall *call, void *user_data)
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static int create_interface(struct supplicant_task *task)
@@ -496,6 +799,8 @@ static void remove_interface_reply(DBusPendingCall *call, void *user_data)
 	free_task(task);
 
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static int remove_interface(struct supplicant_task *task)
@@ -505,6 +810,9 @@ static int remove_interface(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
+	if (task->path == NULL)
+		return 0;
+
 #if 0
 	if (task->created == FALSE) {
 		connman_device_set_powered(task->device, FALSE);
@@ -644,7 +952,7 @@ static int remove_network(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
-	if (task->netpath == NULL)
+	if (task->netpath == NULL || task->path == NULL)
 		return -EINVAL;
 
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
@@ -729,6 +1037,9 @@ static int disconnect_network(struct supplicant_task *task)
 
 	DBG("task %p", task);
 
+	if (task->path == NULL)
+		return -EINVAL;
+
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, task->path,
 				SUPPLICANT_INTF ".Interface", "disconnect");
 	if (message == NULL)
@@ -951,7 +1262,8 @@ static int set_network(struct supplicant_task *task,
 				dbus_message_unref(message);
 				return err;
 			}
-		} else if (g_strcmp0(eap, "peap") == 0) {
+		} else if (g_strcmp0(eap, "peap") == 0 ||
+				   g_strcmp0(eap, "ttls") == 0) {
 			int err;
 
 			err = set_network_peap(network, &dict, passphrase);
@@ -984,7 +1296,7 @@ static int set_network(struct supplicant_task *task,
 		const char *auth_alg = "OPEN";
 		const char *key_index = "0";
 
-		if (task->mac80211 == TRUE)
+		if (task->cfg80211 == TRUE)
 			auth_alg = "OPEN SHARED";
 
 		connman_dbus_dict_append_basic(&dict, "auth_alg",
@@ -1011,6 +1323,17 @@ static int set_network(struct supplicant_task *task,
 						"wep_key0", DBUS_TYPE_BYTE,
 							&key, size / 2);
 				free(key);
+			} else if (size == 5 || size == 13) {
+				unsigned char *key = malloc(13);
+				int i;
+				if (key == NULL)
+					size = 0;
+				for (i = 0; i < size; i++)
+					key[i] = (unsigned char) passphrase[i];
+				connman_dbus_dict_append_fixed_array(&dict,
+						"wep_key0", DBUS_TYPE_BYTE,
+								&key, size);
+				free(key);
 			} else
 				connman_dbus_dict_append_basic(&dict,
 						"wep_key0", DBUS_TYPE_STRING,
@@ -1062,8 +1385,6 @@ static void scan_reply(DBusPendingCall *call, void *user_data)
 	task->scan_call = NULL;
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
 		connman_device_set_scanning(task->device, FALSE);
@@ -1075,6 +1396,8 @@ static void scan_reply(DBusPendingCall *call, void *user_data)
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 
@@ -1416,6 +1739,7 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 {
 	struct supplicant_task *task = user_data;
 	struct supplicant_result result;
+	struct supplicant_block *block;
 	struct connman_network *network;
 	DBusMessage *reply;
 	DBusMessageIter array, dict;
@@ -1427,16 +1751,9 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	DBG("task %p", task);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL) {
-		get_properties(task);
-		return;
-	}
 
-	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
-		dbus_message_unref(reply);
-		get_properties(task);
-		return;
-	}
+	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
+		goto unref;
 
 	memset(&result, 0, sizeof(result));
 	result.frequency = -1;
@@ -1514,6 +1831,17 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	if (result.path[0] == '\0')
 		goto done;
 
+	if (result.name) {
+		block = g_hash_table_lookup(task->hidden_blocks, result.name);
+		if (block) {
+			enable_network(task, block->netpath, FALSE);
+			g_hash_table_remove(task->hidden_blocks, block->ssid);
+		}
+	}
+
+	if (result.ssid == NULL)
+		task->hidden_found = TRUE;
+
 	if (result.frequency > 0 && result.frequency < 14)
 		result.frequency = 2407 + (5 * result.frequency);
 	else if (result.frequency == 14)
@@ -1573,7 +1901,8 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	if (result.name != NULL && result.name[0] != '\0')
 		connman_network_set_name(network, result.name);
 
-	connman_network_set_blob(network, "WiFi.SSID",
+	if (result.ssid_len != 0)
+		connman_network_set_blob(network, "WiFi.SSID",
 						result.ssid, result.ssid_len);
 
 	connman_network_set_string(network, "WiFi.Mode", mode);
@@ -1600,8 +1929,11 @@ done:
 	g_free(result.name);
 	g_free(result.ssid);
 
+unref:
 	dbus_message_unref(reply);
 
+	dbus_pending_call_unref(call);
+
 	get_properties(task);
 }
 
@@ -1611,8 +1943,17 @@ static void get_properties(struct supplicant_task *task)
 	char *path;
 
 	path = g_slist_nth_data(task->scan_results, 0);
-	if (path == NULL)
+	if (path == NULL) {
+		if (task->hidden_found == TRUE) {
+			/*
+			 * We're done with regular scanning, let's enable
+			 * the missing network blocks if there are hidden
+			 * SSIDs around.
+			 */
+			hidden_block_enable(task);
+		}
 		goto noscan;
+	}
 
 	message = dbus_message_new_method_call(SUPPLICANT_NAME, path,
 						SUPPLICANT_INTF ".BSSID",
@@ -1666,8 +2007,6 @@ static void scan_results_reply(DBusPendingCall *call, void *user_data)
 	DBG("task %p", task);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		goto noscan;
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
 		goto done;
@@ -1697,10 +2036,14 @@ static void scan_results_reply(DBusPendingCall *call, void *user_data)
 		task->scan_results = g_slist_append(task->scan_results, path);
 	}
 
+	task->hidden_found = FALSE;
+
 	g_strfreev(results);
 
 	dbus_message_unref(reply);
 
+	dbus_pending_call_unref(call);
+
 	get_properties(task);
 
 	return;
@@ -1708,7 +2051,8 @@ static void scan_results_reply(DBusPendingCall *call, void *user_data)
 done:
 	dbus_message_unref(reply);
 
-noscan:
+	dbus_pending_call_unref(call);
+
 	task->result_call = NULL;
 
 	if (task->scanning == TRUE) {
@@ -1784,6 +2128,8 @@ static int task_connect(struct supplicant_task *task)
 	unsigned int ssid_len;
 	int err;
 
+	g_hash_table_foreach(task->hidden_blocks, block_reset, task);
+
 	connman_inet_ifup(task->ifindex);
 
 	address = connman_network_get_string(task->network, "Address");
@@ -1794,10 +2140,11 @@ static int task_connect(struct supplicant_task *task)
 
 	DBG("address %s security %s", address, security);
 
-	if (security == NULL && passphrase == NULL)
+	if (security == NULL)
 		return -EINVAL;
 
-	if (g_str_equal(security, "none") == FALSE && passphrase == NULL)
+	if (passphrase == NULL && g_str_equal(security, "none") == FALSE &&
+				g_str_equal(security, "ieee8021x") == FALSE)
 		return -EINVAL;
 
 	remove_network(task);
@@ -1868,11 +2215,6 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 	if (state == WPA_INVALID)
 		return;
 
-	if (task->scanning == TRUE && state != WPA_SCANNING) {
-		connman_device_cleanup_scanning(task->device);
-		task->scanning = FALSE;
-	}
-
 	prevstate = task->state;
 	task->state = state;
 
@@ -1897,8 +2239,6 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 							bssid, bssid_len);
 
 		/* carrier on */
-		connman_network_set_method(task->network,
-				CONNMAN_IPCONFIG_METHOD_DHCP);
 		connman_network_set_connected(task->network, TRUE);
 		break;
 
@@ -1953,27 +2293,23 @@ badstate:
 							oldstate, newstate);
 }
 
-static DBusHandlerResult supplicant_filter(DBusConnection *conn,
+static gboolean supplicant_filter(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	struct supplicant_task *task;
 	const char *member, *path;
 
-	if (dbus_message_has_interface(msg,
-				SUPPLICANT_INTF ".Interface") == FALSE)
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
-
 	member = dbus_message_get_member(msg);
 	if (member == NULL)
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+		return TRUE;
 
 	path = dbus_message_get_path(msg);
 	if (path == NULL)
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+		return TRUE;
 
 	task = find_task_by_path(path);
 	if (task == NULL)
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+		return TRUE;
 
 	DBG("task %p member %s", task, member);
 
@@ -1984,7 +2320,7 @@ static DBusHandlerResult supplicant_filter(DBusConnection *conn,
 	else if (g_str_equal(member, "StateChange") == TRUE)
 		state_change(task, msg);
 
-	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+	return TRUE;
 }
 
 int supplicant_start(struct connman_device *device)
@@ -2006,8 +2342,8 @@ int supplicant_start(struct connman_device *device)
 		goto failed;
 	}
 
-	task->mac80211 = connman_inet_is_mac80211(task->ifindex);
-	if (task->mac80211 == FALSE)
+	task->cfg80211 = connman_inet_is_cfg80211(task->ifindex);
+	if (task->cfg80211 == FALSE)
 		connman_warn("Enabling quirks for unsupported driver");
 
 	task->range = g_try_malloc0(sizeof(struct iw_range));
@@ -2027,7 +2363,8 @@ int supplicant_start(struct connman_device *device)
 	task->state = WPA_INVALID;
 	task->disconnecting = FALSE;
 	task->pending_network = NULL;
-
+	task->hidden_blocks = g_hash_table_new_full(g_str_hash, g_str_equal,
+							NULL, remove_block);
 	task_list = g_slist_append(task_list, task);
 
 	return create_interface(task);
@@ -2054,6 +2391,7 @@ int supplicant_stop(struct connman_device *device)
 	g_free(task->range);
 
 	task_list = g_slist_remove(task_list, task);
+	g_hash_table_destroy(task->hidden_blocks);
 
 	if (task->scan_call != NULL) {
 		dbus_pending_call_cancel(task->scan_call);
@@ -2215,9 +2553,8 @@ static void supplicant_remove(DBusConnection *conn, void *user_data)
 	}
 }
 
-static const char *supplicant_rule = "type=signal,"
-				"interface=" SUPPLICANT_INTF ".Interface";
 static guint watch;
+static guint iface_watch;
 
 static int supplicant_create(void)
 {
@@ -2230,19 +2567,20 @@ static int supplicant_create(void)
 
 	DBG("connection %p", connection);
 
-	if (dbus_connection_add_filter(connection,
-				supplicant_filter, NULL, NULL) == FALSE) {
-		dbus_connection_unref(connection);
-		connection = NULL;
-		return -EIO;
-	}
-
-	dbus_bus_add_match(connection, supplicant_rule, NULL);
-	dbus_connection_flush(connection);
-
 	watch = g_dbus_add_service_watch(connection, SUPPLICANT_NAME,
 			supplicant_probe, supplicant_remove, NULL, NULL);
 
+	iface_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						SUPPLICANT_INTF ".Interface",
+						NULL, supplicant_filter,
+						NULL, NULL);
+
+	if (watch == 0 || iface_watch == 0) {
+		g_dbus_remove_watch(connection, watch);
+		g_dbus_remove_watch(connection, iface_watch);
+		return -EIO;
+	}
+
 	return 0;
 }
 
@@ -2253,13 +2591,8 @@ static void supplicant_destroy(void)
 
 	DBG("connection %p", connection);
 
-	if (watch > 0)
-		g_dbus_remove_watch(connection, watch);
-
-	dbus_bus_remove_match(connection, supplicant_rule, NULL);
-	dbus_connection_flush(connection);
-
-	dbus_connection_remove_filter(connection, supplicant_filter, NULL);
+	g_dbus_remove_watch(connection, watch);
+	g_dbus_remove_watch(connection, iface_watch);
 
 	dbus_connection_unref(connection);
 	connection = NULL;
