@@ -218,13 +218,20 @@ int __connman_device_enable(struct connman_device *device)
 
 	err = device->driver->enable(device);
 	if (err < 0) {
-		if (err == -EINPROGRESS)
+		if (err == -EINPROGRESS) {
 			device->powered_pending = TRUE;
+			device->offlinemode = FALSE;
+			if (__connman_profile_get_offlinemode() == TRUE)
+				__connman_profile_set_offlinemode(FALSE, FALSE);
+		}
 		return err;
 	}
 
 	device->powered_pending = TRUE;
 	device->powered = TRUE;
+	device->offlinemode = FALSE;
+	if (__connman_profile_get_offlinemode() == TRUE)
+		__connman_profile_set_offlinemode(FALSE, FALSE);
 
 	__connman_technology_enable_device(device);
 
@@ -696,6 +703,8 @@ static void unregister_network(gpointer data)
 
 	connman_element_unregister((struct connman_element *) network);
 
+	__connman_network_set_device(network, NULL);
+
 	connman_network_unref(network);
 }
 
@@ -1340,6 +1349,9 @@ static void set_offlinemode(struct connman_element *element, gpointer user_data)
 
 	device->offlinemode = offlinemode;
 
+	if (device->blocked == TRUE)
+		return;
+
 	powered = (offlinemode == TRUE) ? FALSE : TRUE;
 
 	if (device->powered == powered)
