@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -40,8 +40,7 @@ struct connman_ipv4 {
 	struct in_addr broadcast;
 };
 
-static int set_ipv4(struct connman_element *element,
-			struct connman_ipv4 *ipv4, const char *nameserver)
+static int set_ipv4(struct connman_element *element, struct connman_ipv4 *ipv4)
 {
 	struct ifreq ifr;
 	struct sockaddr_in addr;
@@ -95,11 +94,6 @@ static int set_ipv4(struct connman_element *element,
 
 	close(sk);
 
-	if (nameserver == NULL)
-		connman_error("No nameserver for %s defined", ifr.ifr_name);
-
-	connman_resolver_append(ifr.ifr_name, NULL, nameserver);
-
 	return 0;
 }
 
@@ -125,8 +119,6 @@ static int clear_ipv4(struct connman_element *element)
 
 	DBG("ifname %s", ifr.ifr_name);
 
-	connman_resolver_remove_all(ifr.ifr_name);
-
 	memset(&addr, 0, sizeof(addr));
 	addr.sin_family = AF_INET;
 	addr.sin_addr.s_addr = INADDR_ANY;
@@ -172,10 +164,13 @@ static char *index2name(int index)
 
 static int ipv4_probe(struct connman_element *element)
 {
+	struct connman_service *service;
+	struct connman_ipconfig *ipconfig;
 	struct connman_element *connection;
 	struct connman_ipv4 ipv4;
 	const char *address = NULL, *netmask = NULL, *broadcast = NULL;
-	const char *nameserver = NULL;
+	const char *nameserver = NULL, *pac = NULL;
+	char *timeserver = NULL;
 
 	DBG("element %p name %s", element, element->name);
 
@@ -188,6 +183,10 @@ static int ipv4_probe(struct connman_element *element)
 
 	connman_element_get_value(element,
 			CONNMAN_PROPERTY_ID_IPV4_NAMESERVER, &nameserver);
+	connman_element_get_value(element,
+			CONNMAN_PROPERTY_ID_IPV4_TIMESERVER, &timeserver);
+	connman_element_get_value(element,
+			CONNMAN_PROPERTY_ID_IPV4_PAC, &pac);
 
 	DBG("address %s", address);
 	DBG("netmask %s", netmask);
@@ -199,9 +198,23 @@ static int ipv4_probe(struct connman_element *element)
 	memset(&ipv4, 0, sizeof(ipv4));
 	ipv4.address.s_addr = inet_addr(address);
 	ipv4.netmask.s_addr = inet_addr(netmask);
-	ipv4.broadcast.s_addr = inet_addr(broadcast);
+	if (broadcast)
+		ipv4.broadcast.s_addr = inet_addr(broadcast);
+	else
+		ipv4.broadcast.s_addr = ipv4.address.s_addr |
+						~ipv4.netmask.s_addr;
 
-	set_ipv4(element, &ipv4, nameserver);
+	set_ipv4(element, &ipv4);
+
+	service = __connman_element_get_service(element);
+
+	if (pac != NULL)
+		__connman_service_set_proxy_autoconfig(service, pac);
+
+	if (nameserver != NULL)
+		__connman_service_append_nameserver(service, nameserver);
+
+	connman_timeserver_append(timeserver);
 
 	connection = connman_element_create(NULL);
 
@@ -209,6 +222,11 @@ static int ipv4_probe(struct connman_element *element)
 	connection->index   = element->index;
 	connection->devname = index2name(element->index);
 
+	ipconfig = __connman_service_get_ipconfig(service);
+	if (ipconfig != NULL)
+		__connman_ipconfig_set_element_ipv6_gateway(
+						ipconfig, connection);
+
 	if (connman_element_register(connection, element) < 0)
 		connman_element_unref(connection);
 
@@ -217,8 +235,25 @@ static int ipv4_probe(struct connman_element *element)
 
 static void ipv4_remove(struct connman_element *element)
 {
+	const char *nameserver = NULL;
+	char *timeserver = NULL;
+
 	DBG("element %p name %s", element, element->name);
 
+	connman_element_get_value(element,
+			CONNMAN_PROPERTY_ID_IPV4_NAMESERVER, &nameserver);
+	connman_element_get_value(element,
+			CONNMAN_PROPERTY_ID_IPV4_TIMESERVER, &timeserver);
+
+	connman_timeserver_remove(timeserver);
+
+	if (nameserver != NULL) {
+		struct connman_service *service;
+
+		service = __connman_element_get_service(element);
+		__connman_service_remove_nameserver(service, nameserver);
+	}
+
 	clear_ipv4(element);
 }
 
